"""Pit-viper offers configuration management capabilities.

This package supports accessing parameters from environment variables,
and configuration files.

Usage:
    >>> from pit import viper
    >>> from pathlib import Path
    >>>
    >>> viper.auto_env()
    >>>
    >>> viper.set_config_path(Path() / "config")
    >>> viper.set_config_name("my_config")
    >>> viper.set_config_type("toml")
    >>> viper.set("foo", "default-value")
    >>>
    >>> viper.load_config()
    >>> foo = viper.get("foo")

For more information, see the https://github.com/leoschleier/pit-viper.
"""
from pit.viper.config import get_conf as get
from pit.viper.config import (
    load_config,
    set_config_name,
    set_config_path,
    set_config_type,
)
from pit.viper.config import set_conf as set  # noqa: A001
from pit.viper.env import auto_env

__all__ = [
    "auto_env",
    "get",
    "load_config",
    "set",
    "set_config_name",
    "set_config_path",
    "set_config_type",
]



"""Load environment variables from a `.env` file."""
import os
from pathlib import Path
from typing import Self

_EQUALS = "="
_NEW_LINE = "\n"
_ENCLOSING_CHARS = f" \"'{_NEW_LINE}"

_env_enabled = False


class UnsupportedFileFormatError(Exception):
    """Raised when the file format is not supported."""

    def __init__(self: Self, path: Path) -> None:
        """Initialize the exception."""
        super().__init__(f"Unsupported file format: {path.suffix}")


def auto_env(
    *,
    path: Path | None = None,
    overwrite: bool = False,
) -> dict[str, str]:
    """Populate environment variables with variables from a `.env` file.

    The `.env` file is being read as text using the UTF-8 character
    encoding.

    Parameters
    ----------
    path : Path | None, optional
        Path to the `.env` file. If not set, `pit-viper` will try to
        find the file in the current working directory, by default None
    overwrite : bool, optional
        Whether or not to overwrite existing environment variables with
        the variables from the `.env` file, by default False

    Returns
    -------
    dict[str, str]
        Environment variables after populating.

    Raises
    ------
    UnsupportedFileFormatError
        Raised if the file format is not supported, i.e., if the
        specified path does not lead to a `.env` file.
    """
    if path is None:
        path = Path(".env")
    if path.exists():
        if ".env" in [path.name, path.suffix]:
            _populate_env(path=path, overwrite=overwrite)
        else:
            raise UnsupportedFileFormatError(path=path)

    global _env_enabled  # noqa: PLW0603
    _env_enabled = True

    return dict(os.environ)


def is_env_enabled() -> bool:
    """Check if environment variables are enabled.

    Environment variables are enabled if the `auto_env` function has
    been called.
    """
    return _env_enabled


def _populate_env(*, path: Path, overwrite: bool = False) -> None:
    """Populate environment variables."""
    with path.open(encoding="utf-8") as stream:
        for line in stream:
            if line == _NEW_LINE:
                continue

            k, v = line.split(_EQUALS)

            # Strip any combination of enclosing characters.
            k = k.strip(_ENCLOSING_CHARS)
            v = v.strip(_ENCLOSING_CHARS)

            _set_variable(key=k, value=v, overwrite=overwrite)


def _set_variable(key: str, value: str, *, overwrite: bool = False) -> None:
    """Set an environment variable."""
    if overwrite:
        os.environ[key] = value
    else:
        os.environ.setdefault(key, value)



"""Internal IO functions for viper."""
import json
import tomllib
from pathlib import Path
from typing import Any

import yaml


def read_config(path: Path) -> dict[str, Any]:
    """Read a config file.

    This function supports configs from JSON, TOML, and YAML files.

    Parameters
    ----------
    path : Path
       Path to the config file.

    Returns
    -------
    dict[str, Any]
        Config

    Raises
    ------
    ValueError
        Raised if the config file extension is not supported.
    """
    file_extension = path.suffix if path.suffix else path.name

    match file_extension:
        case ".json":
            return _read_json(path)
        case ".toml":
            return _read_toml(path)
        case ".yml" | ".yaml":
            return _read_yaml(path)
        case _:
            msg = f"Config file extension {file_extension} not supported."
            raise ValueError(msg)


def _read_json(path: Path) -> dict[str, Any]:
    """Read a JSON file.

    Parameters
    ----------
    path : Path
        Path to the JSON file.

    Returns
    -------
    dict[str, Any]
        JSON file contents.
    """
    with path.open("rb") as stream:
        return json.load(stream)


def _read_toml(path: Path) -> dict[str, Any]:
    """Read a TOML file.

    Parameters
    ----------
    path : Path
        Path to the TOML file.

    Returns
    -------
    dict[str, Any]
        TOML file contents.
    """
    with path.open("rb") as stream:
        return tomllib.load(stream)


def _read_yaml(path: Path) -> dict[str, Any]:
    """Read a YAML file.

    Parameters
    ----------
    path : Path
        Path to the YAML file.

    Returns
    -------
    dict[str, Any]
        YAML file contents.
    """
    with path.open("rb") as stream:
        return yaml.safe_load(stream)



"""Load config from file."""
import os
from pathlib import Path
from typing import Any

from pit.viper import _io, env

_config_path: Path | None = None
_config_name: str = ""
_config_type: str = ""
_config: dict[str, Any] = {}


def set_config_name(name: str) -> None:
    """Set the config name.

    Parameters
    ----------
    name : str
        Name of the config file.
    """
    global _config_name  # noqa: PLW0603
    _config_name = name


def set_config_path(path: Path) -> None:
    """Set the config path.

    Parameters
    ----------
    path : Path
        Path to the config file or the directory containing the config
        file.
    """
    global _config_path  # noqa: PLW0603
    _config_path = path


def set_config_type(type_: str) -> None:
    """Set the config type.

    Supported config formats: json, toml, yaml, yml.

    Parameters
    ----------
    type_ : str
        Config format.
    """
    global _config_type  # noqa: PLW0603
    _config_type = type_.lower()


def get_conf(key: str, default: Any = None) -> Any:
    """Get a value from the config.

    Parameters
    ----------
    key : str
        Key to get the value for.
    default : Any, optional
        Default value to return if the key is not found, by default None

    Returns
    -------
    Any
        Value for the key.
    """
    if env.is_env_enabled():
        env_value = os.environ.get(key)
        if env_value is not None:
            return env_value

    keys = key.split(".")
    config_value = _config

    for k in keys:
        if isinstance(config_value, dict):
            config_value = config_value.get(k, default)

    return config_value


def set_conf(key: str, value: Any) -> None:
    """Set a value in the config.

    Parameters
    ----------
    key : str
        Key to set the value for.
    value : Any
        Value to set.
    """
    keys = key.split(".")
    config = _config

    for k in keys[:-1]:
        if isinstance(config, dict):
            config = config.setdefault(k, {})

    config[keys[-1]] = value


def load_config() -> None:
    """Load the config from the config file.

    Raises
    ------
    FileNotFoundError
        Raised if the config path is not set or the path does not exist.
    """
    if not _config_path:
        msg = "Config path not set."
        raise FileNotFoundError(msg)

    config_file = (
        f"{_config_name}.{_config_type}" if _config_type else _config_name
    )
    config_path = _config_path / config_file

    config = _io.read_config(config_path)
    for key, value in config.items():
        _safe_set(key, value)


def _safe_set(key: str, value: Any) -> None:
    """Set a value in the config without overwriting nested defaults.

    Parameters
    ----------
    key : str
        Key to set the value for.
    value : Any
        Value to set.
    """
    if isinstance(value, dict):
        for (
            k,  # pyright: ignore [reportUnknownVariableType]
            v,  # pyright: ignore [reportUnknownVariableType]
        ) in value.items():
            _safe_set(f"{key}.{k}", v)
    else:
        set_conf(key, value)
